
#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "windows.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 2,6,0,2502
 PRODUCTVERSION 2,6,0,2502
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", "Cisco OpenH264 codec"
            VALUE "CompanyName", "Cisco Systems Inc."
            VALUE "FileDescription", "Cisco OpenH264 codec"
            VALUE "FileVersion", "2.6.0.2502"
            VALUE "InternalName", "openh264.dll"
            VALUE "LegalCopyright", "© 2011-2015 Cisco and/or its affiliates. All rights reserved."
            VALUE "OriginalFilename", "openh264.dll"
            VALUE "ProductName", "Cisco OpenH264 codec"
            VALUE "ProductVersion", "2.6.0.2502"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////
